/**
  Physics simulator demo program.
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2011-02-23 (Public Domain)
*/
#include "physics/world.h"
#include "ogl/glsl.h"
#include "soil/SOIL.h"

#include "ogl/gpgpu.h" /* gpu_env libraries: framebuffers, textures, etc */

/* LAME! just include library bodies here, for easy linking */
#include "physics/world.cpp" 
#include "physics/config.cpp"
#include "ogl/glsl.cpp"
#include "soil/SOIL.c"
#include "soil/stb_image_aug.c"

float fract(float x) {
	return x-floor(x);
}

vec3 fract(const vec3 &src) {
	return vec3(fract(src.x),fract(src.y),fract(src.z));
}

std::string s(double val) { /* convert a float to a string */
	char buf[100];
	sprintf(buf,"%f",val);
	return buf;
}


/* Random float between 0 and 1 */
float randf(void) {
        float f=rand();
        return f/RAND_MAX;
}

/* Random vector where x y and z vary between 0 and 1 */
vec3 randVec3(void) {
        return vec3(randf(),randf(),randf());
}



/* An object that can push or pull on particles. */
class particlePusher {
public:
	virtual void draw(physics::library &lib) =0;
};



/* A sphere-shaped pusher */
class spherePusher : public particlePusher {
public:
	vec3 location;
	float radius;
	vec4 color;
	spherePusher() {
		location=vec3(0.0);
		radius=1.0;
		color=vec4(1,0,0,0.5);
	}
	
	virtual void push(const vec3 &direction) { /* ignore push */ }
	
	virtual void draw(physics::library &lib)
	{
		int slices=8,stacks=5;
		for (int pass=1;pass<=3;pass++) {
			glPushMatrix();
			if (pass==3) { glScalef(1,1,0); /* squish shadows onto ground */ }
			glTranslatef(VEC3_TO_XYZ(location));
			
			if (pass!=2) glDepthMask(GL_FALSE); /* transparent -> no depth writes */
			if (pass==1) { /* first pass: transparent outline */
				glColor4fv(color);
				glutSolidSphere(radius,slices,stacks);
			} 
			else if (pass==2) { /* second pass: wireframe */
				glColor4f(0,0,0,1.0);
				glLineWidth(1.0);
				glutWireSphere(radius,slices,stacks);
			} 
			else if (pass==3 && location.z>=0.0) { /* final pass: shadow */
				glColor4f(0,0,0,0.5);
				glutWireSphere(radius,slices,stacks);
				glColor4f(0,0,0,0.3);
				glutSolidSphere(radius,slices,stacks);
			}
			glPopMatrix();
			glDepthMask(GL_TRUE); /* back to normal */
		}
	}
};


/* A force-based particle-repelling sphere */
class softSphere : public spherePusher {
public:
	softSphere(vec3 location_,float radius_,vec4 color_=vec4(1,0,0,0.5)) {
		location=location_; radius=radius_; color=color_;
	}
};

/* A constraint-based rolling ground sphere */
class rollySphere : public spherePusher {
public:
	int x,y; /* pixel coordinates of the particle I am connected to */
	rollySphere(int x_,int y_,vec4 color_,gpu_array &pos_tex) {
		x=x_; y=y_;
		pos_tex.bind_fbo();
		glReadPixels(x,y,1,1, GL_RGB,GL_FLOAT,(float *)location); // read our position
		color=color_;
	}
	
	virtual void push(const vec3 &direction) 
	{ /* transmit push down to connected node? */
		location+=direction; 
		if (location.z<radius) location.z=radius; /* don't go belowground */
	}
};

/* A particle system */
class particlesystem : public physics::object {
	gpu_env env; /* used for all gpgpu work */
	
	int w,h; /* dimensions of cloth texture: w*h total particles */
	
	/* Per-particle data is stored in these arrays: */
	gpu_array pos_tex, vel_tex, tmp_tex;
	
	unsigned int cur_sphere; /* the sphere we're currently controlling with the camera */
	std::vector<spherePusher *> spheres;
public:
	particlesystem(void) 
		:physics::object(0.001),  /* <- our timestep, in seconds */
		env(true), // already initialized OpenGL
		w(130),h(w), // size of GPGPU textures (width by height)
		pos_tex(env,"pos_tex",w,h,0,GL_RGB32F_ARB), // position (m)
		vel_tex(env,"vel_tex",w,h,0,GL_RGB32F_ARB), // velocity (m/s)
		tmp_tex(env,"tmp_tex",w,h,0,GL_RGB32F_ARB)
	{
		//  Because I don't have an interface for "uniform" variables,
		//   I'm setting up all the parameters here.
		env.utils+="const float dt="+s(timestep.dt)+";\n"; 
		
		float mesh_size=10.0; // size of whole cloth, in meters
		env.utils+="const float mesh_size="+s(mesh_size)+";\n"; 
		
		float edge_length=mesh_size/w; // length of each edge
		env.utils+="const float edge_length="+s(edge_length)+";\n"; 
		env.utils+="const float edge_length_diagonal="+s(edge_length*sqrt(2.0))+";\n"; 
		
		float fabric_density=3.0; // Kg per square meter of fabric
		float mass=edge_length*edge_length*fabric_density;
		env.utils+="const float mass="+s(mass)+";\n";  // Kg per node
		
	// Create the initial particle attributes
		GPU_RUN(pos_tex,
			gl_FragColor = vec4(
				(location.x-0.5)*mesh_size, // X
				(location.y-0.5)*mesh_size, // Y
				6.0, // Z
				1.0 /* alpha */
			)
		);
		
		GPU_RUN(vel_tex,gl_FragColor=vec4(-0.2,-0.2,0.2,1.0) );
		
	/* Make the control spheres */
		spherePusher *s;
		
		s=new softSphere(vec3(0.0),1.0,vec4(0.0,1.0,0.0,0.1)); // green
		spheres.push_back(s);
		env.add_uniform3fv("fearlocation",s->location);
	
		s=new rollySphere(1,h-1,vec4(1.0,0.0,0.0,0.2),pos_tex); // red
		spheres.push_back(s);
		
		s=new rollySphere(w-2,h-2,vec4(1.0,0.0,1.0,0.2),pos_tex); // purple
		spheres.push_back(s);
		
		cur_sphere=0;
	
		gpgpu_reset();
		
		std::cout<<"Final run parameters: \n"<<env.utils;
	}
	
	void simulate(physics::library &lib) {
		float dt=timestep.dt;
		
		// Fix the position of the camera-controlled sphere
		vec3 camera_to_sphere=-camera_orient.z*3.0-camera_orient.y*0.4;
		if (!lib.key_toggle['f']) { /* imposing a camera controlled sphere */
			spheres[0]->location=camera+camera_to_sphere;
		}
		
	// Loop over spheres, simulate and resolve collisions
		for (unsigned int i=0;i<spheres.size();i++) 
		{ 
		  for (unsigned int j=i+1;j<spheres.size();j++) {
			vec3 along=spheres[i]->location - spheres[j]->location;
			float len=length(along);
			float minlen=spheres[i]->radius+spheres[j]->radius;
			if (len<minlen) 
			{ /* these spheres are intersecting */
				spheres[i]->push( along*(minlen-len));
				spheres[j]->push(-along*(minlen-len));
			}
		  }
		}
	
	// Start timestep: update positions
		GPU_RUN(tmp_tex,
			vec3 pos=vec3(texture2D(pos_tex,location));
			vec3 vel=vec3(texture2D(vel_tex,location));
			
			if (location.y<0.95 || abs(location.x-0.5)<0.4) /* clamp far corners */
				pos+=dt*vel;
			
			/* Check position constraints: 
			     the ground is small nonzero because zero causes an explosion (!?) on ATI.
			*/
			if (pos.z<=0.000001) {pos.z=0.000001;} // hit the ground
			
			/* Check user interface sphere */
			vec3 fear=pos-fearlocation;
			if (length(fear)<=1.0) { // move away from the sphere
				pos=normalize(fear)+fearlocation;
			}
			
			gl_FragColor = vec4(pos,1.0);
		);
		tmp_tex.swapwith(pos_tex); // <- can't read and write pos directly...
		
	// Use new positions to calculate an updated net force,
	//    then use that net force to update velocities (all in one pass!)
		GPU_RUNMAIN(tmp_tex,
		/* Return the position of the node at these texture coordinates */
		vec3 pos_at(vec2 loc) { return vec3(texture2D(pos_tex,loc)); }
		
		/* Return the net force along the edge between these two nodes */
		vec3 edge_force(vec3 spos,vec3 epos,float rest_length)
		{
			vec3 along=epos-spos;
			float len=length(along);
			
			if (len<=0.0) return vec3(0.0); /* free boundaries--otherwise get NaNs! */
			
			const float spring_constant=1.0e3; /* N force per m displacement */
			vec3 pull=(len-rest_length)*spring_constant*normalize(along);
			return pull; /* gets added to net force */
		}
		
		/* Return the bend-resisting force pushing the center node back */
		vec3 bend_force(vec3 left,vec3 center,vec3 right)
		{
			vec3 nalong=right-center;
			vec3 along=left-center;
			float lengths=length(along)*length(nalong);
			
			if (lengths<=0.0) return vec3(0.0); /* free boundaries--otherwise get NaNs! */
			
			float dotalong=dot(along,nalong)/lengths;
			float bendiness=max(1.0+dotalong,0.5);
			const float bend_constant=1.0e2; /* N force per cos of bending (try zero!) */
			vec3 bendforce=bendiness*bend_constant*(nalong+along);
			return bendforce;
		}
		
		/* Return the color of this GLSL pixel */
		void main(void) {
			vec3 pos=pos_at(location);
			vec3 vel=vec3(texture2D(vel_tex,location));
			
			/* Calculate net force from all surrounding edges */
			vec2 dx=vec2(pos_texScale.x,0.0); vec2 dy=vec2(0.0,pos_texScale.y);
			
			vec3 L=pos_at(location-dx); vec3 R=pos_at(location+dx); // left, right
			vec3 B=pos_at(location-dy); vec3 T=pos_at(location+dy); // bottom, top
			
			vec3 netforce=
				+edge_force(pos,L,edge_length) // X
				+edge_force(pos,R,edge_length)
				
				+edge_force(pos,T,edge_length) // Y
				+edge_force(pos,B,edge_length)
				
				+edge_force(pos,pos_at(location+dx+dy),edge_length_diagonal) // diagonals
				+edge_force(pos,pos_at(location-dx-dy),edge_length_diagonal)
				
				+edge_force(pos,pos_at(location+dx-dy),edge_length_diagonal) // other diagonals
				+edge_force(pos,pos_at(location-dx+dy),edge_length_diagonal)
			;
			
			if (true) {
			/* Add in bending resistance force, for smoother fabric */
			vec3 L2=pos_at(location-2.0*dx); vec3 R2=pos_at(location+2.0*dx); // 2-away neighbors
			vec3 B2=pos_at(location-2.0*dy); vec3 T2=pos_at(location+2.0*dy); //   (for bending)
			netforce +=	
				+bend_force(L,pos,R) +bend_force(T,pos,B) // bending (direct force)
				-0.5*( // bending (reaction force from each neighbor) 
					+bend_force(L2,L,pos) +bend_force(pos,R,R2)
					+bend_force(B2,B,pos) +bend_force(pos,T,T2)
				)
			;
			}
			
			/* F = m A, so A = F/m */
			vec3 acc=netforce/mass; 

			acc.z+=-9.8;  /* gravity */
			
			/* "wind drag": moderate */
			acc+=-3.2*vel;
			vel+=dt*acc;

			if (pos.z<=0.001) { /* hit the ground (velocity side) */
			  vel.z=0.9*abs(vel.z); /* velocity must point upward */
			  vel*=0.8; /* ground drag */
			}
			
			/* user interface sphere (velocity side checks) */
			vec3 fear=pos-fearlocation;
			if (length(fear)<=1.0) { // moving along the sphere
				vel*=0.8; /* sphere drag */
			}
			
			gl_FragColor = vec4(vel,1.0);
		}
		);
		tmp_tex.swapwith(vel_tex); // <- can't read and write vel directly...
		
		// For debugging, print coordinates at a corner:
		if (0) {
			vec3 v[1];
			vel_tex.read(v[0],1,1); printf("vel[0]=%f,%f,%f\n",v[0].x,v[0].y,v[0].z);
			pos_tex.read(v[0],1,1); printf("pos[0]=%f,%f,%f\n",v[0].x,v[0].y,v[0].z);
		}
		
		gpgpu_reset(); // back to onscreen rendering
	}
	
	
	void drawSpheres(physics::library &lib) {

		// Draw all spheres
		for (unsigned int i=0;i<spheres.size();i++)
			spheres[i]->draw(lib);
	}
	
#if 0 // Render one big vertex buffer object (100% GPU, but vertex texture fetch fails on pre-DirectX 10 ATI cards)

	void draw(physics::library &lib) {
	/* Prepare a vertex buffer object with (x,y) texture coordinates in a triangle strip */
		static int nverts;
		static GLuint vbo=0;
		if (vbo==0) 
		{
			typedef std::pair<short,short> vec2i; /* has two ints */
			std::vector<vec2i > vbuf; /* data */
			
			for (int y=1;y+2<h;y++) {
				vbuf.push_back(vec2i(1,y));
				int x;
				for (x=1;x+1<w;x++) {
					vbuf.push_back(vec2i(x,y));
					vbuf.push_back(vec2i(x,y+1));
				}
				vbuf.push_back(vec2i(x-1,y+1));
			}
			glEnd();
			nverts=vbuf.size();
			
			glGenBuffersARB(1,&vbo);
			glBindBufferARB(GL_ARRAY_BUFFER_ARB,vbo);
			glBufferDataARB(GL_ARRAY_BUFFER_ARB,
				sizeof(vbuf[0])*vbuf.size(),&vbuf[0],GL_STATIC_DRAW_ARB);
        	
			glEnableClientState(GL_VERTEX_ARRAY);
			glVertexPointer(2,GL_SHORT,sizeof(vbuf[0]), (void *)0); /* position is first thing in each struct */
		}
		
	/* Make GLSL to render the fabric directly from the pos_tex texture */
		static GLhandleARB prog=MAKE_PROGRAM_OBJECT3(
			/* Vertex program */
			
			/* Return the position of the node at these texture coordinates */
			vec3 pos_at(vec2 loc) { return vec3(texture2D(pos_tex,loc)); }
			
			void main(void) {
				color=gl_Color;
				texcoords=pos_texScale*(vec2(gl_Vertex)+vec2(0.5)); // from pixels
				location=pos_at(texcoords);
				
				vec2 dx=vec2(pos_texScale.x,0.0); vec2 dy=vec2(0.0,pos_texScale.y);
				vec3 L=pos_at(texcoords-dx); vec3 R=pos_at(texcoords+dx); // left, right
				vec3 B=pos_at(texcoords-dy); vec3 T=pos_at(texcoords+dy); // bottom, top
				normal=normalize(cross(L-R,T-B));
				
				if (color.r==0.0) location.z=0.0; /* shadow */
				gl_Position = gl_ModelViewProjectionMatrix*vec4(location,1.0);
			}
		,
			/* Shared code */
			varying vec4 color;
			varying vec3 normal;
			varying vec3 location;
			varying vec2 texcoords;
			uniform sampler2D pos_tex;
			uniform vec2 pos_texScale; /* scale from pixel to pos_tex coords */
		,
			uniform vec3 camera;
			uniform sampler2D banner_tex;
			/* Fragment program */
			void main(void) {
				vec3 toCamera=camera-location;
				vec3 n=normal;
				if (dot(toCamera,n)<0.0) n=-n;
			
				vec3 lighting;
				float white=dot(n,normalize(vec3(0.2,-0.5,1.0)));	
				if (white<0.0) white=-0.3*white; // little bit of reflected light	
				lighting=vec3(0.9)*(0.1+white)
					+0.1*vec3(n.x,n.y,n.z);
				//lighting=normal; // color by normal
				vec4 texture=vec4(vec3(texture2D(banner_tex,4.0*texcoords)),1.0);
				gl_FragColor=color*texture*vec4(lighting,1.0);
			}
		);
		
		/* Set up the uniform variables for the shader */
		glUseProgramObjectARB(prog);
		static GLuint texture=SOIL_load_OGL_texture(
			"banner.jpg",0,0,
			SOIL_FLAG_MIPMAPS|SOIL_FLAG_TEXTURE_REPEATS|SOIL_FLAG_INVERT_Y
		);
		glActiveTexture(GL_TEXTURE1); /* Texture unit 1: the banner */
		glBindTexture(GL_TEXTURE_2D,texture);
		glFastUniform1i(prog,"banner_tex",1);
		
		pos_tex.bind_tex(0); /* Texture unit 0: XYZ positions */
		glFastUniform1i(prog,"pos_tex",0);
		
		glFastUniform2fv(prog,"pos_texScale",1,vec3(1.0/w,1.0/h,0.0));
		glFastUniform3fv(prog,"camera",1,camera);
		
		glBindBufferARB(GL_ARRAY_BUFFER_ARB,vbo);
	
	/* Draw shadows */
		glDepthMask(GL_FALSE); /* avoid Z fighting in shadows */
		glColor4f(0,0,0,0.7); /* draw shadows */
		glDrawArrays(GL_TRIANGLE_STRIP,0,nverts);
		glDepthMask(GL_TRUE);
	
	/* Draw main fabric (finally!) */
		glColor4f(1,1,1,1); 
		glDrawArrays(GL_TRIANGLE_STRIP,0,nverts);
		
	/* Back to normal rendering */
		glBindBufferARB(GL_ARRAY_BUFFER_ARB,0);
		glUseProgramObjectARB(0);
	
	/* Draw everything else */
		drawSpheres(lib);
	}
	
#else // Pull back particle positions, for rendering from the CPU side
	std::vector<vec3> particles;
	
	// Lighting utility functions
	const vec3 &pos(int x,int y) const { 
		return particles[y*w+x]; 
	}
	vec3 color(int x,int y) const { 
		vec3 xaxis=pos(x+1,y)-pos(x-1,y);
		vec3 yaxis=pos(x,y+1)-pos(x,y-1);
		vec3 normal=normalize(cross(xaxis,yaxis));
		
		vec3 toCamera=camera-pos(x,y); /* toward camera */
		if (dot(normal,toCamera)<0.0f) normal=-normal; /* flip to face camera */
		
		//return normalize(normal);  // color by normal: bright but pretty!
		static vec3 lightDirection=normalize(vec3(0.0,-0.4,1.0));
		float white=dot(normal,lightDirection);
		if (white<0.0) white=-0.3*white; /* facing down: a little reflected light */
		return vec3(white+normal.x*0.1,white+fabs(normal.x)*0.05,white+normal.y*0.1);
	}
	
	void draw(physics::library &lib) {
		int n=w;

		particles.resize(w*h);
		pos_tex.read(particles[0],w,h);
		
		gpgpu_reset(); // back to onscreen rendering
		
		printf("Particles[0]=%f,%f,%f\n",
			particles[0].x,particles[0].y,particles[0].z);
	
		// Draw the shadow (basically just the 2D version)
		glDepthMask(GL_FALSE); // avoid Z buffer fighting
		glColor4f(0,0,0,0.7);
		glBegin(GL_TRIANGLE_STRIP);
		for (int y=1;y+2<n;y++) {
			glVertex2fv(pos(1,y));
			int x;
			for (x=1;x+1<n;x++) {
				glVertex2fv(pos(x,y));
				glVertex2fv(pos(x,y+1));
			}
			glVertex2fv(pos(x-1,y+1));
		}
		glEnd();
		glDepthMask(GL_TRUE);
		
		static GLuint texture=SOIL_load_OGL_texture(
			"banner.jpg",0,0,
			SOIL_FLAG_MIPMAPS|SOIL_FLAG_TEXTURE_REPEATS|SOIL_FLAG_INVERT_Y
		);
		glBindTexture(GL_TEXTURE_2D,texture);
		glMatrixMode(GL_TEXTURE);
		glLoadIdentity();
		glScalef(4.0/w,4.0/h,0.0); /* from simulated pixel, to texture coordinates */
		
		// Draw the fabric, with texture
		glEnable(GL_TEXTURE_2D);
		glBegin(GL_TRIANGLE_STRIP);
		for (int y=1;y+2<n;y++) {
			glTexCoord2i(1,y);
			glVertex3fv(pos(1,y));
			int x;
			for (x=1;x+1<n;x++) {
				glTexCoord2i(x,y);
				glColor3fv(color(x,y));
				glVertex3fv(pos(x,y));
				
				glTexCoord2i(x,y+1);
				glColor3fv(color(x,y+1));
				glVertex3fv(pos(x,y+1));
			}
			glTexCoord2i(x-1,y+1);
			glVertex3fv(pos(x-1,y+1));
		}
		glEnd();
		
		glLoadIdentity();
		glDisable(GL_TEXTURE_2D);
		glMatrixMode(GL_MODELVIEW);
		
		drawSpheres(lib);
	}
#endif /* which draw */
};


/* Called create a new simulation */
void physics_setup(physics::library &lib) {
	if (!lib.key_toggle['f']) {
		//camera=vec3(0.0,-5.0,1.7);
		camera_orient.x=vec3(1,0,0);
		//camera_orient.z=normalize(-vec3(0,0.6,0.4));
	}
	lib.world->add(new particlesystem());
	lib.background[0]=lib.background[1]=0.6;
	lib.background[2]=1.0; // light blue
}

/* Called every frame to draw everything */
void physics_draw_frame(physics::library &lib) {
	/* use SOIL ground texture */
	static GLuint texture=SOIL_load_OGL_texture(
		"ground.jpg",0,0,SOIL_FLAG_MIPMAPS|SOIL_FLAG_TEXTURE_REPEATS
	);
	glBindTexture(GL_TEXTURE_2D,texture);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAX_ANISOTROPY_EXT,32);
	/* Draw opaque white ground plane */
	int groundsize=500;
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);
	glColor3f(1,1,1);
	glPushMatrix();glTranslatef(0,0,-0.01); /* <- shove down, to reduce Z fighting */
	glBegin(GL_QUADS);
	glTexCoord2i(-groundsize,-groundsize); glVertex2i(-groundsize,-groundsize);
	glTexCoord2i(+groundsize,-groundsize); glVertex2i(+groundsize,-groundsize);
	glTexCoord2i(+groundsize,+groundsize); glVertex2i(+groundsize,+groundsize);
	glTexCoord2i(-groundsize,+groundsize); glVertex2i(-groundsize,+groundsize);
	glEnd();
	glPopMatrix();
	glDisable(GL_TEXTURE_2D);

	/* Point and line smoothing needs normal alpha blending */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	/* Now draw objects */
	lib.world->draw(lib);
}


